/*utmpctl.c by Samuel Behan <behan@frida.fri.utc.sk> (c) 2001
*/
#include <stdio.h>
#include <stdlib.h>
#include <utmp.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

/*header*/
#include "utmpctl.h"

#ifdef DEBUG
#define say(msg) fprintf(stderr, msg"\n");
#else
#define say(msg)
#endif

/*utmp types*/
static char utmp_types[10][15] =
{
    "UT_UNKNOWN", "RUN_LEVEL", "BOOT_TIME", "NEW_TIME", "OLD_TIME",
    "INIT_PROCESS", "LOGIN_PROCESS", "USER_PROCESS", "DEAD_PROCESS",
    "ACCOUNTING" 
};

/*config*/
#define UTSESS_DEFAULT	0001
#define UTSESS_LIST	0002
#define UTSESS_MODIFY	0004
#define UTSESS_ADD	0010
#define UTSESS_DEL	0020
#define UTSESS_INVIS	0040
#define UTSESS_UNINVIS	0100
#define UTSESS_WIPE	0200
struct Config {
    struct utmp *rec;		/*record modifs*/
    char *file;			/*utmp filename*/
    int action;			/*action type*/
    int pos;			/*position of rec. to modify*/
};

char *utmp_type(int type)
{
    if (type < 0 || type > 9)
	return "INCORRECT";
    return utmp_types[type];
}

void record_header()
{
    printf("%4s|%-15s|%6s|%6s|%-3s|%8s|%-10s|%-15s\n",
	       "Pos", " Type", "PID", "Device", "ID", "Username", "Hostname", "Time");
    printf("--------------------------------------------------------------------------------\n");
    return;
}

#define utmp_FORMAT "%4d|%-15.15s|%6d|%-6s|%-3.3s|%-8.8s|%-10.10s|%15s\n"
void record_print(num, rec)
int num;
struct utmp *rec;
{    
    if(strncmp(rec->ut_id, "si", 2) == 0)
	printf(utmp_FORMAT, num, "", 0, "", rec->ut_id, "", "", "");
    else
    {   char curr_time[25] = "";
    	strftime(curr_time, 25, "%a %d %b %T", localtime(&rec->ut_time));
	printf(utmp_FORMAT, num, utmp_type(rec->ut_type), rec->ut_pid, rec->ut_line,
	       rec->ut_id, rec->ut_user, rec->ut_host, curr_time); }
    return;
}


int main(int argc, char **argv)
{
    int c = 0, recs = 0;
    struct Config *config;
    struct utmp *utmp_rec;
    extern char *optarg;
    
    /*init utmp rec*/
    config	= (struct Config *)malloc(sizeof(struct Config));
    memset(config, 0, sizeof(struct Config));
    config->rec	= (struct utmp *)malloc(sizeof(struct utmp));
    memset(config->rec, 0, sizeof(struct utmp));
    config->rec->ut_type= -1;
    config->file	= _PATH_UTMP;
    config->pos		= -1;

    /*parse command line*/
    while ((c = getopt(argc, argv, "-T:P:D:I:U:H:f:hmlaiuw")) != EOF) {
	switch (c) {
	    case 1:		/*default  - number */
	    {   char *endptr[1];
		endptr[0]	= malloc(sizeof(char)*strlen(optarg));
		config->action	= UTSESS_MODIFY;
	        config->pos	= strtol(optarg, (char **)endptr, 10);
		if(optarg[0] == '\0' || (endptr != NULL && endptr[0][0] != '\0')
				|| config->rec->ut_type > 10)
		{   fprintf(stderr, "position argument not a number '%s'\n", optarg);
		    exit(EXIT_FAILURE); }
		}break;
	    case 'T':		/*modifier - TYPE */
	    {   char *endptr[1];
/*FIXME: some reord types are add only !!!*/
		endptr[0]	= malloc(sizeof(char)*strlen(optarg));
		config->rec->ut_type	= strtol(optarg, (char **)endptr, 10);
		if(optarg[0] == '\0' || (endptr != NULL && endptr[0][0] != '\0')
				|| config->rec->ut_type > 10)
		{   fprintf(stderr, "unknown utmp type '%s'\n", optarg);
		    exit(EXIT_FAILURE); }
		}break;
	    case 'P':		/*modifier - PID */
	        config->rec->ut_pid	= (pid_t)atoi(optarg);
		break;
	    case 'D':		/*modifier - DEVICE */
	        strncpy(config->rec->ut_line, optarg, UT_LINESIZE);
		config->rec->ut_line[UT_LINESIZE-1]	= '\0';
		break;
	    case 'I':		/*modifier - INIT ID */
	        strncpy(config->rec->ut_id, optarg, 4);
		config->rec->ut_id[4-1]	= '\0';
		break;
	    case 'U':		/*modifier - USERNAME */
	        strncpy(config->rec->ut_user, optarg, UT_NAMESIZE);
		config->rec->ut_user[UT_NAMESIZE-1]	= '\0';
		break;
	    case 'H':		/*modifier - HOSTNAME */
	        strncpy(config->rec->ut_host, optarg, UT_HOSTSIZE);
		config->rec->ut_host[UT_HOSTSIZE-1]	= '\0';
		break;
	    case 'f':		/*utmp file*/
	    	config->file	= optarg;
	    	break;
	    case 'l':
	    	config->action	|= UTSESS_LIST;
		break;
	    case 'w':
	    	config->action	|= UTSESS_WIPE;
		break;
	    case 'i':		/*session invisible*/
	        config->action	&= ~(UTSESS_UNINVIS);
		config->action	|= UTSESS_INVIS;
		break;
	    case 'u':		/*session un-invisible*/
	        config->action	&= ~(UTSESS_INVIS);
		config->action	|= UTSESS_UNINVIS;
		break;
	    case 'm':
	    	config->action	&= ~(UTSESS_ADD);
	    	config->action	|= UTSESS_MODIFY;
		break;
	    case 'a':		/*session add */
	        config->action	&= ~(UTSESS_MODIFY);
	    	config->action	= UTSESS_ADD;
		break;
	    case 'h':		/*HELP */
		fprintf(stderr, APP_USAGE, argv[0]);
		fprintf(stderr, APP_VERSION);
		fprintf(stderr, APP_HELP);
		exit(0);
	    default:		/*error */
		exit(1);
	}
    }
    if(config->action & UTSESS_MODIFY && config->rec->ut_id[0] == 0
    		&& config->pos == -1)
    {   fprintf(stderr, "no matching record to modify\n");
    	exit(EXIT_FAILURE); }
    
    /*set utmp file path*/
    if((c = open(config->file, O_RDONLY)) == -1)
    {	fprintf(stderr, "error opening utmp file: %s\n", strerror(errno));
        exit(errno); }
    else close(c);
    utmpname(config->file);
    setutent();			/*rewind utmp even if no need to */
    /*do perform actions*/
    if(config->action & UTSESS_MODIFY)
    {	if(config->rec->ut_id[0] == '\0' && config->pos != -1)
	{   while((utmp_rec = getutent()) != NULL
		  && recs++ < config->pos);
	    if(utmp_rec	== NULL)
            {    fprintf(stderr, "no such utmp record %d (use -a for add)\n", config->pos);
		 exit(EXIT_FAILURE); }
	    if(config->rec->ut_type == -1)
		 config->rec->ut_type = utmp_rec->ut_type;
	    strcpy(config->rec->ut_id, utmp_rec->ut_id); }
	else if((utmp_rec = getutid(config->rec)) == NULL)
	{    fprintf(stderr, "no such utmp record %d (use -a for add)\n", config->pos);
	     exit(EXIT_FAILURE); }
	 /*get old values*/
	 if(!(config->action & UTSESS_WIPE))
	 {   if(config->rec->ut_line[0] == '\0')
	       	 memmove(config->rec->ut_line, utmp_rec->ut_line, sizeof(char)*UT_LINESIZE);
	     if(config->rec->ut_pid == 0)
	       	 config->rec->ut_pid	= utmp_rec->ut_pid;
	     if(config->rec->ut_user[0] == '\0')
	       	 memmove(config->rec->ut_user, utmp_rec->ut_user, sizeof(char)*UT_NAMESIZE);
	     if(config->rec->ut_host[0] == '\0')
	       	 memmove(config->rec->ut_host, utmp_rec->ut_host, sizeof(char)*UT_HOSTSIZE);
	     memmove(&config->rec->ut_exit, &utmp_rec->ut_exit, sizeof(struct exit_status));
	     config->rec->ut_session	= utmp_rec->ut_session;
	     memmove(&config->rec->ut_tv, &utmp_rec->ut_tv, sizeof(struct timeval));
	     memmove(&config->rec->ut_addr_v6, &utmp_rec->ut_addr_v6, sizeof(int32_t)*4);
	     memmove(&config->rec->ut_tv, &utmp_rec->ut_tv, sizeof(struct timeval));
/*	     memmove(&config->rec->pad, &utmp_rec->pad, sizeof(char)*20);*/
         }
	 /*update utmp record*/
	 setutent();
	 pututline(config->rec); 
	 endutent(); }
    if(config->action & UTSESS_ADD)
    {  while((utmp_rec = getutent()) != NULL);
       pututline(config->rec);
       endutent(); }
    if(config->action & UTSESS_LIST || config->action == 0)
    {  setutent();
       record_header();
       recs	= 0;
       while ((utmp_rec = getutent()) != NULL)
           record_print(recs++, utmp_rec); }
    endutent();			/*close utmp */
    return 0;
}
